@@ -40,6 +40,7 @@
 # EXE_SUFFIX        -- [in, opt] '.exe' on Windows/Cygwin/similar
 # VERSION_MAJOR     -- [in] sys.version_info.major
 # VERSION_MINOR     -- [in] sys.version_info.minor
+# ABI_THREAD        -- [in] either 't' for free-threaded builds or ''
 # PYWINVER          -- [in] the Windows platform-specific version (e.g. 3.8-32)
 
 # ** Values read from the environment **
@@ -172,17 +173,18 @@
 # ******************************************************************************
 
 platlibdir = config.get('platlibdir') or PLATLIBDIR
+ABI_THREAD = ABI_THREAD or ''
 
 if os_name == 'posix' or os_name == 'darwin':
     BUILDDIR_TXT = 'pybuilddir.txt'
     BUILD_LANDMARK = 'Modules/Setup.local'
     DEFAULT_PROGRAM_NAME = f'python{VERSION_MAJOR}'
-    STDLIB_SUBDIR = f'{platlibdir}/python{VERSION_MAJOR}.{VERSION_MINOR}'
+    STDLIB_SUBDIR = f'{platlibdir}/python{VERSION_MAJOR}.{VERSION_MINOR}{ABI_THREAD}'
     STDLIB_LANDMARKS = [f'{STDLIB_SUBDIR}/os.py', f'{STDLIB_SUBDIR}/os.pyc']
-    PLATSTDLIB_LANDMARK = f'{platlibdir}/python{VERSION_MAJOR}.{VERSION_MINOR}/lib-dynload'
+    PLATSTDLIB_LANDMARK = f'{platlibdir}/python{VERSION_MAJOR}.{VERSION_MINOR}{ABI_THREAD}/lib-dynload'
     BUILDSTDLIB_LANDMARKS = ['Lib/os.py']
     VENV_LANDMARK = 'pyvenv.cfg'
-    ZIP_LANDMARK = f'{platlibdir}/python{VERSION_MAJOR}{VERSION_MINOR}.zip'
+    ZIP_LANDMARK = f'{platlibdir}/python{VERSION_MAJOR}{VERSION_MINOR}{ABI_THREAD}.zip'
     DELIM = ':'
     SEP = '/'
 
@@ -229,17 +231,16 @@ use_environment = config.get('use_environment', 1)
 
 pythonpath = config.get('module_search_paths')
 pythonpath_was_set = config.get('module_search_paths_set')
+stdlib_dir = config.get('stdlib_dir')
+stdlib_dir_was_set_in_config = bool(stdlib_dir)
 
 real_executable_dir = None
-stdlib_dir = None
 platstdlib_dir = None
 
 # ******************************************************************************
 # CALCULATE program_name
 # ******************************************************************************
 
-program_name_was_set = bool(program_name)
-
 if not program_name:
     try:
         program_name = config.get('orig_argv', [])[0]
@@ -311,7 +312,10 @@ if ENV_PYTHONEXECUTABLE or ENV___PYVENV_LAUNCHER__:
         # and should not affect base_executable.
         base_executable = f"{dirname(library)}/bin/python{VERSION_MAJOR}.{VERSION_MINOR}"
     else:
-        base_executable = executable
+        # Use the real executable as our base, or argv[0] otherwise
+        # (on Windows, argv[0] is likely to be ENV___PYVENV_LAUNCHER__; on
+        # other platforms, real_executable is likely to be empty)
+        base_executable = real_executable or executable
 
     if not real_executable:
         real_executable = base_executable
@@ -409,13 +413,14 @@ if not base_executable:
 if not real_executable:
     real_executable = base_executable
 
-try:
-    real_executable = realpath(real_executable)
-except OSError as ex:
-    # Only warn if the file actually exists and was unresolvable
-    # Otherwise users who specify a fake executable may get spurious warnings.
-    if isfile(real_executable):
-        warn(f'Failed to find real location of {base_executable}')
+if real_executable:
+    try:
+        real_executable = realpath(real_executable)
+    except OSError as ex:
+        # Only warn if the file actually exists and was unresolvable
+        # Otherwise users who specify a fake executable may get spurious warnings.
+        if isfile(real_executable):
+            warn(f'Failed to find real location of {base_executable}')
 
 if not executable_dir and os_name == 'darwin' and library:
     # QUIRK: macOS checks adjacent to its library early
@@ -428,12 +433,12 @@ if not executable_dir and os_name == 'darwin' and library:
 
 # If we do not have the executable's directory, we can calculate it.
 # This is the directory used to find prefix/exec_prefix if necessary.
-if not executable_dir:
+if not executable_dir and real_executable:
     executable_dir = real_executable_dir = dirname(real_executable)
 
 # If we do not have the real executable's directory, we calculate it.
 # This is the directory used to detect build layouts.
-if not real_executable_dir:
+if not real_executable_dir and real_executable:
     real_executable_dir = dirname(real_executable)
 
 # ******************************************************************************
@@ -509,11 +514,12 @@ if ((not home_was_set and real_executable_dir and not py_setpath)
             build_stdlib_prefix = build_prefix
         else:
             build_stdlib_prefix = search_up(build_prefix, *BUILDSTDLIB_LANDMARKS)
-        # Always use the build prefix for stdlib
-        if build_stdlib_prefix:
-            stdlib_dir = joinpath(build_stdlib_prefix, 'Lib')
-        else:
-            stdlib_dir = joinpath(build_prefix, 'Lib')
+        # Use the build prefix for stdlib when not explicitly set
+        if not stdlib_dir_was_set_in_config:
+            if build_stdlib_prefix:
+                stdlib_dir = joinpath(build_stdlib_prefix, 'Lib')
+            else:
+                stdlib_dir = joinpath(build_prefix, 'Lib')
         # Only use the build prefix for prefix if it hasn't already been set
         if not prefix:
             prefix = build_stdlib_prefix
@@ -545,8 +551,9 @@ else:
         prefix, had_delim, exec_prefix = home.partition(DELIM)
         if not had_delim:
             exec_prefix = prefix
-        # Reset the standard library directory if it was already set
-        stdlib_dir = None
+        # Reset the standard library directory if it was not explicitly set
+        if not stdlib_dir_was_set_in_config:
+            stdlib_dir = None
 
 
     # First try to detect prefix by looking alongside our runtime library, if known
@@ -562,7 +569,8 @@ else:
         if STDLIB_SUBDIR and STDLIB_LANDMARKS and not prefix:
             if any(isfile(joinpath(library_dir, f)) for f in STDLIB_LANDMARKS):
                 prefix = library_dir
-                stdlib_dir = joinpath(prefix, STDLIB_SUBDIR)
+                if not stdlib_dir_was_set_in_config:
+                    stdlib_dir = joinpath(prefix, STDLIB_SUBDIR)
 
 
     # Detect prefix by looking for zip file
@@ -573,7 +581,7 @@ else:
                 prefix = executable_dir
         else:
             prefix = search_up(executable_dir, ZIP_LANDMARK)
-        if prefix:
+        if prefix and not stdlib_dir_was_set_in_config:
             stdlib_dir = joinpath(prefix, STDLIB_SUBDIR)
             if not isdir(stdlib_dir):
                 stdlib_dir = None
